/***************************************************************************
                          rttydemodulator.cpp  -  description
                             -------------------
    begin                : Mon Jun 4 2001
    copyright            : (C) 2001 by Volker Schroer
    email                : dl1ksv@gmx.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *    based on the work of  Moe Wheatly, AE4JY                             *  
 ***************************************************************************/

#include "rttydemodulator.h"

RTTYDemodulator::RTTYDemodulator():CDemodulator()
{
ShiftOn = false;
NumberOfStopBits = 2;
parity = 0;

NxSamples = 0;
twiddles = NULL;
bins = NULL;
history = NULL;
NumberofTones = 2;
FrequencyNumber = NULL;
// Disable PhaseDisplay

}
RTTYDemodulator::~RTTYDemodulator()
{
if (twiddles != NULL)
  delete twiddles;
if (bins != NULL)
	delete bins;
if (history != NULL )
	delete history;
if (FrequencyNumber != NULL)
	delete FrequencyNumber;
}
/** returns the asci char corresponding to the baudot code */
char RTTYDemodulator::baudot_code(char data)
{
/** Table of letters */

static const char letters[32] =	{'<','E','\n','A',' ','S','I','U',
								 '\n','D','R','J','N','F','C','K',
								 'T','Z','L','W','H','Y','P','Q',
								 'O','B','G','^','M','X','V','^'};


/** Table of symbols */
static const char symbols[32] = {'<','3','\n','-',' ',',','8','7',
 '\n','$','4','#',',','',':','(',
 '5','+',')','2','','6','0','1',
'9','7','','^','.','/','=','^'};

char c;

	switch (data)
	 {
		case 0x1f :
				ShiftOn = false;  //LTRS
				c = 0;
				break;
		case 0x1b :
		 		ShiftOn = true;   //FIGS
		 		c = 0;
		  	break;
		default:	if (!ShiftOn)
								c=letters[data];
							 else
							 c=symbols[data];
						 break;
 		}


if ( c == ' ')				// Unshift on Space
	ShiftOn =false;	

return c;
}

bool RTTYDemodulator::Init(double FS,int NumberOfSamples)
{

int i;
float x;

zaehler=0;

Baudrate=45;
NumberOfBits=5;
NxSamples=NumberOfSamples;
SamplesPerBit=int (FS/Baudrate+0.5);
Status = WaitingForStart;
actSample=0;
data=0;
BitsInData=0;
bitcounter=0;
c=0;
count=0;

FrequencyChanged = false;


ave1=0.0;
ave2=0.0;

y0ges=0.0;
y1ges=0.0;


if ( (twiddles = new float_complex [SamplesPerBit]) == NULL )
	return false;
if ((bins = new float_complex[NumberofTones]) == NULL )
	return false;
if ( (history = new float[SamplesPerBit] ) == NULL )
	return false;
if ( (FrequencyNumber=new int[NumberofTones]) == NULL )
	return false;
ptr = 0;
FrequencyNumber[0] = 22 ; // 1000 Hz
FrequencyNumber[1] = 26 ; // 1170 Hz

// Initialize the coefficients for the slfft

for (i=0;i < SamplesPerBit; i++)
{
x =2*i * M_PI*Baudrate/FS;
twiddles[i] = float_complex( cos(x), sin(x) ) ;
history[i]= 0.;
}
for(i=0;i<NumberofTones;i++)
	bins[i]= float_complex(0.,0.);
return true;
}

void RTTYDemodulator::ProcessInput(double *input)
{





char c1;

int x2;
bool bit;

DataAvailable = true;	// Just read the data
actSample = 0;				// Start processing, counter will be incremented while processing
											// the next value
data=input;	
while (DataAvailable)
	{
		switch (Status)
			{
				case WaitingForStart:
					while ( DataAvailable)
						{
							x2=ProcessNextSample();
								if ( x1 > 0 && x2 < 0)
									{
										Status  = CountingStartBits;
										bitcounter = SamplesPerBit/2;
										break;
									}
								else
									x1 = x2;
							}					// End While
							
							break;				
							
				case CountingStartBits:
					while ( DataAvailable  )
							if ( --bitcounter > 0)
								x1 = ProcessNextSample();
							else
								break;		// End while
					if ( bitcounter == 0 ) // 0.5 bits processed
						{
						if ( x1 > 0 )
							{
								Status  = WaitingForStart; // Was'nt the correct start bit
								x2 = x1;
							}	
						else
							{		
								Status = CollectingByte;
								bitcounter = 5;
                CalcQuality();
							}
						}
						
					break;				// We've reached the end of the available data or completed Status

				case Synch:
					while (DataAvailable && (bitcounter >0) )
						{
              bitcounter--;
							if ( synchronize() )
								{
								 Status = CountingStartBits;
	               bitcounter = SamplesPerBit/2;
              	 break;
								}	
						}
						if (bitcounter == 0)
							{
							Status = WaitingForStart;
							x1=1;
							}
					  break;
				case CollectingByte:
					while( DataAvailable && (bitcounter > 0) )
						{
							bit=get_next_bit();							
							if ( DataAvailable )
								{
									CalcQuality();

									if (bit)
										BitsInData++;
									bitcounter--;
									c >>=1;
									if (bit)
										c |= 128;
								}									
							else
								break;			// No more data to be processed at the moment	
							
						}
					if ( bitcounter == 0 ) // Data 'byte' completed
						{
							c >>= (8- NumberOfBits);
							Status = CheckingParity;
							c1=c;
							c1 = baudot_code(c1);  // FIGS or LTRS result in c1 = 0 !
							if ( c1 > 0 )
								emit newSymbol( c1 );
						}	
							
					break;	
					
				case CheckingParity:

					if (parity)
						{
							bit=get_next_bit();
							if (	DataAvailable )
								{
									CalcQuality();

									if (bit)
										BitsInData++;
									if (parity == 1 ) // even
										{
										if ( BitsInData & 1)	// error						
											{
											emit newSymbol('<');
											emit newSymbol('P');
											emit newSymbol('>');
											}
										}
									else							// odd
										{
										if (! (BitsInData & 1))	// error						
											{
											emit newSymbol('<');
											emit newSymbol('P');
											emit newSymbol('>');
											}
										}
									
									}
									else
										break;
								}	
								Status =WaitingForStopBits;
								BitsInData = 0;
//								bitcounter=NumberOfStopBits;
								bitcounter =1;
							
				break;	
									
				case WaitingForStopBits:
					while ( DataAvailable && bitcounter >0 )
						{
						bit = get_next_bit();
						if ( DataAvailable)		// We've got a bit
							{
								CalcQuality();
								bitcounter --;
								if  (!bit)		// Framing Error
									{
									emit newSymbol('<');
									emit newSymbol('F');
									emit newSymbol('>');
      						Status = CollectingByte;   // Lets try whether this was already the start bit
									bitcounter=NumberOfBits;
									c=0;
			            break;
									}
								else
							{
										
								c=0;
								Status=Synch;
//								Status=WaitingForStart;

								bitcounter =  SamplesPerBit/2;
//							Status = ThrowHalfBit;
								break;
							}	

							}
						}
				
				break;
				
				case ThrowHalfBit:
					while ( DataAvailable && --bitcounter > 0)
						x1 = ProcessNextSample();
					if ( bitcounter == 0 )
						Status = WaitingForStart;
				break;			
						
		}	// End switch	
	}	// End While
}

int RTTYDemodulator::ProcessNextSample()
{
float x;
if ( actSample<NxSamples )
	{
	zaehler++;
	DataAvailable = true;
	x= (float) data[actSample];
	actSample++;
  slfft(x);
  y0 = abs(bins[0]);
	y1 = abs(bins[1]);
	if ( y0 > y1 )
		return -1;
	else
 return 1;
	}
else
	DataAvailable = false;
	
return 0;		
}



bool RTTYDemodulator::get_next_bit()
{

int x;


while (DataAvailable)
	{
  x=ProcessNextSample();
	count++;

	if (count < SamplesPerBit )
		{
			if ( count > SamplesPerBit/2)
//			if ( count > 10)

				{

					y0ges +=y0;
					y1ges +=y1;
				
				}
		}
		else
		{
			if ( y0ges > y1ges)
				x= -1;
			else
				x= 1;
			count=0;
			y0ges=0.;
			y1ges=0.;
			break;
		}
	}
return ( (x > 0 ) ? true : false) ;	
}

void RTTYDemodulator::slfft(float x)
{
float old;

old = x - history[ptr];
history[ptr] = x;
ptr = (ptr +1) % SamplesPerBit;

for (int i=0; i <NumberofTones;i++)
	{
	bins[i] = bins[i] + old;
	bins[i] = bins[i] *twiddles [ FrequencyNumber[i]];
	}
}

void RTTYDemodulator::setRxFrequency(double freq)
{
int i;
if ( freq != RxFrequency)
	{
		RxFrequency=freq;
		for(i=0; i < NumberofTones; i++)
			bins[i]=float_complex(0.0,0.0);
		FrequencyNumber[0]=(int) (RxFrequency/Baudrate);
		FrequencyNumber[1]=(int) ((RxFrequency+170.)/Baudrate);
		ptr=0;
		for(i=0;i < SamplesPerBit; i++)	
			history[i]=0.;
		Status = WaitingForStart;									
   }
}

void RTTYDemodulator::CalcQuality()
{
ave2=ave1;
//ave1=0.8*ave1 + 0.2 * abs(y0-y1)/(y0+y1);
ave1=0.9*ave1 + 0.1 * abs(y0-y1)/(y0+y1);
emit setSquelchValue((int)(100*ave1));
}


bool RTTYDemodulator::synchronize()
{
int x;


x=ProcessNextSample();

if (  x <0 )
	return true;
else
return false;

}
